//* -*- mode: c++; c-basic-offset: 4; -*- */
#include "frame_driver.hh"
#include "loc_list.hh"
#include "chan_list.hh"
#include "segdb.hh"

#include "framecpp/FrameCPP.hh"
#include "framecpp/Dimension.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrAdcData.hh"
#include "framecpp/FrProcData.hh"
#include "framecpp/FrRawData.hh"
#include "framecpp/FrSimData.hh"
#include "framecpp/FrStatData.hh"
#include "framecpp/FrVect.hh"
#include <cmath>
#include <unistd.h>

//======================================  Define pointer types
#define FCPP_SHARED_PTRS 1
#include "framecpp/GPSTime.hh"
#if FRAMECPP_USES_LDAS_TOOLS_SHARED_PTR
#include "ldastoolsal/SharedPtr.hh"
#define FRAMCPP_SHARED_PTR(a,b) typedef LDASTools::AL::SharedPtr<FrameCPP::a>     b;
#else
#include <boost/shared_ptr.hpp>
#define FRAMCPP_SHARED_PTR(a,b) typedef boost::shared_ptr<FrameCPP::a>     b;
#endif

FRAMCPP_SHARED_PTR(FrameH,frame_pointer);
FRAMCPP_SHARED_PTR(FrRawData,frrawdata_pointer);
FRAMCPP_SHARED_PTR(FrAdcData,fradcdata_pointer);
FRAMCPP_SHARED_PTR(FrProcData,frprocdata_pointer);
FRAMCPP_SHARED_PTR(FrSimData,frsimdata_pointer);
FRAMCPP_SHARED_PTR(Common::FrStatData,frstatdata_pointer);
FRAMCPP_SHARED_PTR(FrVect,frvect_pointer);

#undef FRAMECPP_SHARED_PTR

using FrameCPP::FrVect;
using FrameCPP::FrameH;
using FrameCPP::FrAdcData;
using FrameCPP::FrRawData;
using FrameCPP::FrProcData;
using FrameCPP::FrSimData;
typedef FrRawData::const_firstAdc_iterator AdcDataIterator;
typedef FrameH::const_procData_iterator    ProcDataIterator;
typedef FrameH::const_simData_iterator     SimDataIterator;
using FrameCPP::FrStatData;
using FrameCPP::GPSTime;
typedef FrStatData::Query stat_query;

using namespace std;
using namespace sends;

//======================================  Constructor
frame_driver::frame_driver(void) {
    FrameCPP::Initialize();
}

//======================================  Destructor
frame_driver::~frame_driver(void) {
}

//======================================  Inline stat
FrStatData&
statRef(frstatdata_pointer& p) {
    return dynamic_cast<FrStatData&>(*p);
}

//======================================  Read frame
int
frame_driver::read_frame(frame_stream& frdr, segment_vect& segv, 
			 int nFrame, gps_type gps) const {

    //----------------------------------  Read in the frame header.
    unsigned int rflgs = 0;
    frame_pointer frame = frdr.ReadFrameH(nFrame, rflgs);
    frame->SetRawData(frrawdata_pointer(new FrRawData));
    gps_type start_frame = frame->GetGTime().GetSeconds();
    if (!start_frame) {
	if (!gps) throw runtime_error("Invalid input frame (start time is 0)");
	cerr << "Frame time stamp is invalid (zero)!, use " << gps << endl;
	start_frame = gps;
    }
    gps_type end_frame   = start_frame + gps_type(frame->GetDt());
    if (end_frame <= start_frame) {
	throw runtime_error("Frame duration is invalid (<=zero)!");
    }

    //----------------------------------  Loop over channels in list
    int rc = 0;
    index_type N = segv.size();
    for (index_type i=0; i<N; ++i) {
	seg_index iseg = segv[i];
	if (Segment_List.is_null(iseg)) continue;

	segment& segi(Segment_List[iseg]);
	gps_type start_fill = segi.getEndGPS();
	if (end_frame <= start_fill) continue;

	gps_type end_fill   = segi.getReserveGPS();
	if (start_frame >= end_fill) continue;

	chan_index ichn  = segi.chan();
	string     cname = Channel_List[ichn].name();
	data_type  typ   = Channel_List[ichn].dtype();
	double     rate  = Channel_List[ichn].rate();

	//------------------------------  Set up to read a channel
	if (debug() >= 2) cerr << "    Read channel: " << cname << endl;

	frvect_pointer pv;

	//------------------------------  Handle static data.
	if (Channel_List[ichn].type() == chan_type::kStatic) {
	    string det  = cname.substr(0, cname.find(":"));
	    if (debug() >= 3) {
		cout << "Search for static channel: " << cname 
		     << " in detector: " << det << endl;
	    }

	    //--------------------------  Read the FrStatData structure.
	    try {
		stat_query statQuery;
		frdr.ReadFrStatData(cname, det, GPSTime(start_fill, 0), 
				    GPSTime(end_fill, 0),
				    stat_query::LATEST_VERSION, statQuery);
		int N = statQuery.size();
		for (int i=0; i<N; ++i) {
		    frstatdata_pointer s = statQuery[i];
		    if (debug() >= 3) {
			cout << "Channel: " << statRef(s).GetName() 
			     << " valid times: " << statRef(s).GetTimeStart() 
			     << "-" << statRef(s).GetTimeEnd() << endl;
		    }
		    if (cname != statRef(s).GetName()) continue;
		    if (start_fill >= statRef(s).GetTimeStart() && 
			start_fill < statRef(s).GetTimeEnd()) {
			pv = statRef(s).RefData()[0];
		    }
		}
	    } catch (exception& e) {
		if (debug()) cerr << "Exception reading FrStat data for "
				  << cname << ": " << e.what() << endl;
	    }

	    //--------------------------  Stat data not available
	    if (!pv) {
		cerr << "Static channel: " << cname 
		     << " not in source: " << name() << endl;
		release_segment(iseg);
		segv[i] = Segment_List.end();
		rc = -1;
		continue;
	    }
	}

	//------------------------------  Handle simulated data.
	if (Channel_List[ichn].type() == chan_type::kSimData) {
	    string det  = cname.substr(0, cname.find(":"));
	    if (debug() >= 3) {
		cout << "Search for static channel: " << cname 
		     << " in detector: " << det << endl;
	    }

	    //--------------------------  Read the FrStatData structure.
	    try {
		FrameH::simData_type& sCntr = frame->RefSimData();
		SimDataIterator simIter = sCntr.begin();
		simIter = sCntr.find(cname, simIter);
		if (simIter != sCntr.end()) {
		    pv = (*simIter)->RefData().front();
		} else {
		    frsimdata_pointer sim(frdr.ReadFrSimData(nFrame, cname));
		    if (sim) {
			sCntr.append(sim);
			pv = sim->RefData().front();
		    }
		}
	    } catch (exception& e) {
		if (debug()) cerr << "Exception reading FrSimData for "
				  << cname << ": " << e.what() << endl;
	    }

	    //--------------------------  Stat data not available
	    if (!pv) {
		cerr << "Simulated channel: " << cname 
		     << " not in source: " << name() << endl;
		release_segment(iseg);
		segv[i] = Segment_List.end();
		rc = -1;
		continue;
	    }
	}

	//------------------------------  Look for a raw (FrAdcData) channel
	frrawdata_pointer raw = frame->GetRawData();
	if (!pv && raw && Channel_List[ichn].type() != chan_type::kReduced) {
	    try {
		AdcDataIterator adcIter = raw->RefFirstAdc().begin();
		adcIter = raw->RefFirstAdc().find(cname, adcIter);
		if (adcIter != raw->RefFirstAdc().end()) {
		    if ((*adcIter)->RefData().empty()) 
			throw runtime_error("FrAdcData has no vector!");
		    pv = (*adcIter)->RefData().front();
		} else {
		    fradcdata_pointer padc=frdr.ReadFrAdcData(nFrame, cname);
		    if (padc) {
			raw->RefFirstAdc().append(padc);
			if (padc->RefData().empty()) 
			    throw runtime_error("FrAdcData has no vector!");
			pv = padc->RefData().front();
			segi.set_units(padc->GetBias(), padc->GetSlope(),
				       padc->GetUnits());
		    }
		}
	    } catch (exception& e) {
		if (debug()) cerr << "Exception reading FrAdcData for "
				  << cname << ": " << e.what() << endl;
	    }
	}

	//------------------------------  Look for FrProcData if no FrAdcData
	if (!pv) {
	    try {
		FrameH::procData_type& pCntr = frame->RefProcData();
		ProcDataIterator procIter = pCntr.begin();
		procIter = pCntr.find(cname, procIter);
		if (procIter != pCntr.end()) {
		    pv = (*procIter)->RefData().front();
		} else {
		    frprocdata_pointer proc(frdr.ReadFrProcData(nFrame, cname));
		    if (proc) {
			pCntr.append(proc);
			pv = proc->RefData().front();
		    }
		}
	    } catch (exception& e) {
		if (debug()) cerr << "Exception reading FrProcData for "
				  << cname << ": " << e.what() << endl;
	    }
	}
	if (!pv) {
	    cerr << "Channel: " + cname + " not in source: " + name() << endl;
	    release_segment(iseg);
	    segv[i] = Segment_List.end();
	    rc = -1;
	    continue;
	}
	pv->Uncompress();

	//------------------------------  Verify sample rate and channel
	//                                data type are as expected. Allow
	//                                it to be wrong to single precision.
	FrameCPP::Dimension d = pv->GetDim(0);
	if (fabs(d.GetDx()*rate - 1.0) > 1e-05) {
	    cerr << "Unexpected sample rate (" << 1/d.GetDx()
		 << ") for channel: " << cname << endl;
	    release_segment(iseg);
	    segv[i] = Segment_List.end();
	    rc = -1;
	    continue;
	}

	data_type fr_dtyp;
	fr_dtyp.set_frame_type(pv->GetType());
	if (typ != fr_dtyp) {
	    cerr << "Unexpected frame data type (" << pv->GetType()
		 << ") for channel: " << cname << endl;
	    release_segment(iseg);
	    segv[i] = Segment_List.end();
	    rc = -1;
	    continue;
	}
	
	//------------------------------  Set units from FrProcData, etc.
	if (segi.units().empty() && !pv->GetUnitY().empty()) {
            segi.set_units(segi.bias(), segi.scale(), pv->GetUnitY());
	}
	
	//------------------------------  Copy data into buffers.
	if (end_frame < end_fill) end_fill = end_frame;
	long istart = long(double(start_fill-start_frame)*rate+0.5);
	long istop  = long(double(end_fill - start_frame)*rate+0.5);
	long length = istop - istart;
	if (debug() >= 3) {
	    cerr << "Copy channel: " << Channel_List[ichn] 
		 << " time: " << start_fill << "-" << end_fill
		 << " length: " << length << " to segment: "  << iseg << endl;
	}
	bool xresult = segi.extend(pv->GetData().get(), istart, length);
	if (xresult) {
	    segi.unuse();
	    segv[i] = Segment_List.end();
	}
    }
    return rc;
}

//======================================  Test whether frame exists
bool 
frame_driver::test_available(const loc_list& ll, gps_type t) const 
{
    frame_loc fl;
    if (!ll.find_loc(t, fl)) return false;
    string file_path = fl.file_path(t, true);
    return (access(file_path.c_str(), R_OK) == 0);
}
